#!/usr/bin/python
# wemust-netdbg v2
# This file accompanies wemust-start.py and jackaudio.py and enables network debug of multiple platforms running wemust.
# It basically provides multiple consoles that grab multicast debug messages from the platforms and print it
#
# depends: GTK3, avahi, dbus (fix)
# works with: python 2.7 (until wemust.py is not adapted also to python3)
#
# ToDo/NiceToHave:
# - add menu bar with connection retry (i.e. restarting the sockets and respawning threads)
# - can we set the network interface from the menu? (risking being platform dependent?)
# - complete integration with SABy (not only SEARCH but also showing in devices list)
# - add a thread listening to incoming ADV messages from SABy (and adding the device to the list)
# - Search by SABy tags is not of interest here, but in the devices list I may filter (this requires implementing SABy tags, SR, NID, etc. in the devices view... maybe a tree to show these properties?)
# - add a Temperature column in the devicesList and print the last temp value there. Maybe add this in a SABy message or another protocol
# - add a colored box for each item in the devicesList for easy hint of which color is related to which device
# - devices not in the deviceList will be white.
#
# Author: Leonardo Gabrielli (l.gabrielli@univpm.it)
# Created: 2014/06/23
# License: GPLv3




WEMUST_NETDBG_VERSION = "0.3.0"

# ---------------- IMPORTS -------------------- #

from gi.repository import Gtk
from gi.repository import GObject as gobject
from gi.repository import Gdk
import threading
import wemust
import socket
import struct
import time, random
import sys, os, string
import subprocess
import re
from collections import deque

# ---------------- PATHS -------------------- #

SRC_PATH=(os.path.split(os.path.realpath(__file__)))[0] + '/'
HELP_FILE_PATH = SRC_PATH + "docs/help.txt"

# ---------------- CMD DEFINES -------------------- #
STARTCMD = "python /usr/bin/jackaudio.py start &"
STOPCMD = "python /usr/bin/jackaudio.py stop &"
REBOOTCMD = "reboot"
HALTCMD = "halt"
ADDRESS_SEPARATOR = ': '
TIMESTAMP_DELIM1 = '['
TIMESTAMP_DELIM2 = ']'

# ---------------- INITS -------------------- #

NCONSOLES = 4
gobject.threads_init()

QUIT_TIMEOUT = 10 #seconds
SCAN_TIMEOUT = 10 #seconds
SOCK_TIMEOUT = 2 #seconds. ALWAYS MAKE SURE SOCK_TIMEOUT << SCAN_TIMEOUT. In my design this is used to allow periodic time cecks in socket blocking operations



# ---------------------------------------------------------- #
# ------------ GLOBAL INTRA-CLASS DATA SHARING ------------- #
# ---------------------------------------------------------- #
class dataShare():
    def __init__(self):
        self.color = ["red", "green", "blue", "brown", "pink", "magenta", "orange", "purple", "yellow"]
        self.CONSOLE_BG_COLOR = Gdk.RGBA(0.9, 0.9, 0.9, 1) # 000 = black, 111 = white. Never use total black or white (black and white fonts are used)
        self.threadsList = []
        self.quitFlag = threading.Event()
        self.printTimestamp = False
        self.printAddress = True
        self.toKill = []
        self.globalQLock = threading.Lock() # one lock for all the queues: reduced parallelism... to change
        self.dbgQ = deque()
        self.jtcliQ = deque()
        self.jtserQ = deque()
        self.jdQ = deque()

    def newRandomColor(self):
        pippo = 1
        # todo


# ------------ UTILITY FUNCS --------------- #

def isValidIP(addr):
    if ( addr ):
        try:
            socket.inet_pton(socket.AF_INET, addr) # AF_INET6 for future IPv6 compat
            return 1
        except socket.error:
            return 0
    else:
        return 0

def isValidHostname(addr):
    if ( addr ):
        return 1 # to implement (some check on irregular chars?)
    else:
        return 0

def isValidHTMLColor(colorStr):
    if ( colorStr ):
        if ( re.match("^#(?:[0-9a-fA-F]{3}){1,2}$", colorStr) != None ):
            return 1
        else:
            return 0



# ---------------------------------------------------------- #
# ---------------- THREAD HANDLER CLASS -------------------- #
# ---------------------------------------------------------- #

# checks for aux threads to kill
class threadHandler(threading.Thread):
    def __init__(self, dataShare):
        threading.Thread.__init__(self)
        self.dataShare = dataShare

    def run(self):
        try:
            while( not self.dataShare.quitFlag.is_set() ):
                if( self.dataShare.toKill == [] ):
                    time.sleep( 2 )
                else:
                    while( self.dataShare.toKill ):
                        print "Killing %s" % self
                        self.dataShare.toKill.pop().join(QUIT_TIMEOUT)

        except (KeyboardInterrupt, SystemExit):
            print("Keyboard Interrupt caught: exiting")
            raise



# ---------------------------------------------------------- #
# ---------------- GUI CLASS FOR NETDBG -------------------- #
# ---------------------------------------------------------- #

class pops: # a tribute to the great Louis Armstrong
    wTree = Gtk.Builder()

    def __init__(self, parent):
        # GUI
        self.builder = Gtk.Builder()
        self.builder.add_from_file(SRC_PATH + "GUI/waitDialog.glade")
        #
        self.waitDialog = self.builder.get_object("waitDialog")
        self.waitDialogBox = self.builder.get_object("waitDialogBox")
        self.waitDialogLabel = self.builder.get_object("waitDialogLabel")
        self.waitDialogSpinner = self.builder.get_object("waitDialogSpinner")
        self.waitDialog.set_modal(True)
        self.waitDialog.set_transient_for(parent)

    def show(self):
        self.waitDialog.show()
        self.waitDialog.set_visible(True)
        self.waitDialogSpinner.start()
        self.waitDialogSpinner.show()
        self.waitDialogSpinner.set_visible(True)

    def hide(self):
        self.waitDialog.hide()


class netdbgGUI:
    wTree = Gtk.Builder()
    clearSema = threading.Semaphore(0)
    scanLock = threading.Lock()
    rawReceivedStuff = []
    STagList = [] # list of SABy tags (not to be confused with tags for the deviceList, which are meant for text highlighting (e.g. coloring)

    def __init__(self, dataShare):
        # inter-class data
        self.dataShare = dataShare
        # GUI
        self.builder = Gtk.Builder()
        self.builder.add_from_file(SRC_PATH + "GUI/wemust-netdbg.glade")
        self.builder.add_from_file(SRC_PATH + "GUI/add-dialog.glade")
        self.builder.add_from_file(SRC_PATH + "GUI/netdbg-about.glade")
        self.builder.add_from_file(SRC_PATH + "GUI/netdbg-help.glade")
        # GUI init
        self.getAll()
        self.connectAll()
        self.initDevicesView()
        # version parse
        file = open(SRC_PATH + 'wemust-connect.py', "r")
        we_co = '-'
        for line in file:
            if re.search('WEMUST_CONNECT_VERSION', line):
                we_co = re.search('\"(.+?)\"', line)
                we_co = we_co.group(1)
        self.aboutDialog.set_version("VERSIONING:\n wemust-connect: v." + we_co + "\n wemust-netdbg: v." + WEMUST_NETDBG_VERSION + "\n wemust.py lib: v." + wemust.WEMUST_GUTS_VERSION)
        # init terminals
        self.termInit()
        dataShare.threadsList.append( threading.Thread(target=self.consoleUpdateThread) )

    # GET ALL GUI OBJECTS FROM THE GLADE FILE
    def getAll(self):
        self.window = self.builder.get_object("netdbgWindow")
        self.staB = self.builder.get_object("startButton")
        self.stoB = self.builder.get_object("stopButton")
        self.rebB = self.builder.get_object("rebootButton")
        self.halB = self.builder.get_object("haltButton")
        self.clAllB = self.builder.get_object("clearAllButton")
        self.addB = self.builder.get_object("addButton")
        self.remB = self.builder.get_object("removeButton")
        self.scanB = self.builder.get_object("scanButton")
        self.diagB1 = self.builder.get_object("defaultDiagnosticButton")
        self.diagB2 = self.builder.get_object("customDiagnosticButton")
        #
        self.diagE = self.builder.get_object("customDiagnosticEntry")
        self.diagE.set_editable(True)
        #self.diagE.set_text("enter bash command")
        self.diagBuf = self.builder.get_object("customDiagnosticBuffer")
        self.diagBufPrev = self.builder.get_object("customDiagnosticBufferPrev") # store previous command
        #
        self.dbgPrintTxt = self.builder.get_object("dbgPrintBuf") # alt: get the GtkTextView and then .get_buffer()
        self.jtcliTxt = self.builder.get_object("jtcliBuf")
        self.jackdTxt = self.builder.get_object("jackdBuf")
        self.jtserTxt = self.builder.get_object("jtserBuf")
        self.textBufList = [self.dbgPrintTxt, self.jtcliTxt, self.jackdTxt, self.jtserTxt]
        #
        self.dbgPrintView = self.builder.get_object("dbgPrintView")
        self.dbgPrintView.override_background_color(Gtk.StateFlags.NORMAL, self.dataShare.CONSOLE_BG_COLOR)
        self.dbgPrintView.override_color(Gtk.StateFlags.NORMAL, Gdk.RGBA(1, 1, 1, 1))
        #
        self.jtcliView = self.builder.get_object("jtcliView")
        self.jtcliView.override_background_color(Gtk.StateFlags.NORMAL, self.dataShare.CONSOLE_BG_COLOR)
        self.jtcliView.override_color(Gtk.StateFlags.NORMAL, Gdk.RGBA(1, 1, 1, 1))
        #
        self.jackdView = self.builder.get_object("jackdView")
        self.jackdView.override_background_color(Gtk.StateFlags.NORMAL, self.dataShare.CONSOLE_BG_COLOR)
        self.jackdView.override_color(Gtk.StateFlags.NORMAL, Gdk.RGBA(1, 1, 1, 1))
        #
        self.jtserView = self.builder.get_object("jtserView")
        self.jtserView.override_background_color(Gtk.StateFlags.NORMAL, self.dataShare.CONSOLE_BG_COLOR)
        self.jtserView.override_color(Gtk.StateFlags.NORMAL, Gdk.RGBA(1, 1, 1, 1))
        #
        self.devicesView = self.builder.get_object("devicesView")
        self.devicesList = self.builder.get_object("devicesList")
        #
        self.statusBar = self.builder.get_object("statusBar")
        self.statusBarCI = self.statusBar.get_context_id("Statusbar")
        #
        self.aboutItem = self.builder.get_object("netdbgAbout")
        self.helpItem = self.builder.get_object("netdbgHelp")


        # ADD DIALOG
        self.addDialog = self.builder.get_object("addDialog")
        self.addDialogButtonBox = self.builder.get_object("addDialogButtonBox")
        self.addDialogApplyButton = self.builder.get_object("addDialogApplyButton")
        self.addDialogCancelButton = self.builder.get_object("addDialogCancelButton")
        self.addEntryIP = self.builder.get_object("addEntryIP")
        self.addEntryIP.set_editable(True)
        self.addEntryHostname = self.builder.get_object("addEntryHostname")
        self.addEntryHostname.set_editable(True)
        self.addEntryColor = self.builder.get_object("addEntryColor")
        self.addEntryColor.set_editable(True)

        # Menus
        self.ViewMenuTogglePrintTimestamp = self.builder.get_object("ViewMenuTogglePrintTimestamp")
        self.ViewMenuTogglePrintAddress = self.builder.get_object("ViewMenuTogglePrintAddress")

        #### FROM about.glade
        self.aboutDialog = self.builder.get_object("aboutDialog")

        #### FROM help.glade
        self.helpWindow = self.builder.get_object("helpWindow")
        self.helpTextView = self.builder.get_object("helpTextView")

    # CONNECT ALL GUI OBJECTS TO SIGNALS
    def connectAll(self):
        self.window.connect ("destroy", properShutdown, self.dataShare)
        self.staB.connect("pressed", self.sendStart)
        self.stoB.connect("pressed", self.sendStop)
        self.rebB.connect("pressed", self.sendReboot)
        self.halB.connect("pressed", self.sendHalt)
        self.clAllB.connect("pressed", self.clearConsoles)
        self.addB.connect("pressed", self.openAddDialog)
        self.remB.connect("pressed", self.remEntry)
        self.scanB.connect("pressed", self.startScan)
        self.diagB1.connect("pressed", self.defaultDiagnostic)
        self.diagB2.connect("pressed", self.diagE_execute, self.diagE)
        self.diagE.connect("activate", self.diagE_execute, self.diagE) # execute entry command
        self.diagE.connect("key-press-event", self.diagE_key_press, self.diagE) # up arrow to recall last command
        self.aboutItem.connect("activate", self.openAbout)
        self.helpItem.connect("activate", self.openHelp)
        self.helpWindow.connect("delete-event", self.closeHelp)
        self.addDialogApplyButton.connect("pressed", self.addDialogApply)
        self.addDialogCancelButton.connect("pressed", self.addDialogCancel)
        self.ViewMenuTogglePrintTimestamp.connect("toggled", self.togglePrintTimestamp)
        self.ViewMenuTogglePrintAddress.connect("toggled", self.togglePrintAddress)

    # POPULATE THE DEVICES VIEW
    def initDevicesView(self):
        # remove these dummy entries once AddButton works and/or scan works
        self.devicesList.append(["dummy...", "0.0.0.0", False, str(self.dataShare.color[1])])
        #self.devicesList.append(["debian-BB2.local", "192.168.23.41", True, str(self.dataShare.color[2])])
        #self.devicesList.append(["debian-BB3.local", "192.168.23.42", False, str(self.dataShare.color[3])])
        #self.devicesList.append(["debian-BB4.local", "192.168.23.43", False, str(color[3])])

        #### tags to colour the text
        for t in self.textBufList:
            t.create_tag("white", foreground="white") # white is for devices not registered in the devicesList
            t.create_tag("black", foreground="black") # black is for devices added manually with bad HTML color
            for i in range(len(self.dataShare.color)):
                t.create_tag(self.dataShare.color[i], foreground=self.dataShare.color[i])


        self.hostname = Gtk.CellRendererText()
        self.ipAddr = Gtk.CellRendererText()
        self.addressIt = Gtk.CellRendererToggle()

        self.column1 = Gtk.TreeViewColumn("hostname", self.hostname, text=0) # text=0 means it takes the string from column 0 of the model
        self.column2 = Gtk.TreeViewColumn("IP", self.ipAddr, text=1)
        self.column3 = Gtk.TreeViewColumn("Address it?", self.addressIt, active=2) # active=2 means it takes the boolean value from column 2 of the model

        self.addressIt.connect("toggled", self.on_toggle)

        self.devicesView.append_column(self.column1)
        self.devicesView.append_column(self.column2)
        self.devicesView.append_column(self.column3)

    # INIT THE TERMINALS
    def termInit(self):
        self.termList = []
        self.dbgT = TermEmu(self.dbgPrintView, self, self.dataShare.dbgQ)
        self.termList.append(self.dbgT)
        self.jtcliT = TermEmu(self.jtcliView, self, self.dataShare.jtcliQ)
        self.termList.append(self.jtcliT)
        self.jackdT = TermEmu(self.jackdView, self, self.dataShare.jdQ)
        self.termList.append(self.jackdT)
        self.jtserT = TermEmu(self.jtserView, self, self.dataShare.jtserQ)
        self.termList.append(self.jtserT)

    # UPDATES THE TEXT ON TERMINALS
    def consoleUpdateThread(self):
        self.sleepTime = 0.005 # idle time in seconds
        while( not self.dataShare.quitFlag.is_set() ):
            try:
                for T in self.termList:
                    T.update()
                time.sleep(self.sleepTime)
            except (KeyboardInterrupt, SystemExit):
                print("Keyboard Interrupt caught: exiting")
                raise

    def togglePrintTimestamp(self, pippo):
        self.dataShare.printTimestamp = not self.dataShare.printTimestamp

    def togglePrintAddress(self, pippo):
        self.dataShare.printAddress = not self.dataShare.printAddress

    def errorPopup(self, pippo, msg):
        dlg = Gtk.MessageDialog(type=Gtk.MESSAGE_ERROR, flags=Gtk.DIALOG_MODAL, buttons=Gtk.BUTTONS_OK, message_format=msg)
        dlg.show()

    def waitPopupShow(self, msg): # shows a GtkSpinner until the semaphore is cleared
        self.waitDialogLabel.set_text(msg)
        self.waitDialogBox.show_all()
        self.waitDialog.show()
        self.waitDialogSpinner.start()


    def waitPopupHide(self):
        # how to get the handle to the spinner and stop it?
        self.waitDialogSpinner.stop()
        self.waitDialog.hide()

    # EXECUTE COMMAND INPUT ON DIAGNOSTIC ENTRY
    def diagE_execute(self, widget, entry):
        if ( entry.get_text_length() != 0): # if empty do nothing
            diagCmd = entry.get_text()
            entry.set_text("")
            if diagCmd:
                self.sendCmd_TCP(diagCmd)
            self.diagBufPrev.set_text(diagCmd,-1) # stores the last command

    # RECALLS LAST COMMAND ON DIAGNOSTIC ENTRY
    def diagE_key_press(self, asd1, event, asd3):
        keypress = Gdk.keyval_name(event.keyval)
        if keypress.lower() == "up": # avoid platform-dependent uppercase/lowercase changes
            self.diagBuf.set_text(self.diagBufPrev.get_text(),-1)

    # TOGGLES THE DEVICE VIEW TOGGLES
    def on_toggle(self, model, path):
        self.devicesList[path][2] = not self.devicesList[path][2]

    # OPEN ABOUT DIALOG
    def openAbout(self, about):
        res = self.aboutDialog.run()
        self.aboutDialog.hide()

    # OPEN HELP WINDOW
    def openHelp(self, help):
        txtFile = open(HELP_FILE_PATH, 'r')
        try:
            helpTxt = txtFile.read()
        finally:
            txtFile.close()
        self.helpTextView.get_buffer().set_text(helpTxt)
        self.helpWindow.show()

    # OPEN ADD DEVICE DIALOG
    def openAddDialog(self, add):
        res = self.addDialog.run()
        self.aboutDialog.hide()

    # CLOSE HELP WINDOW
    def closeHelp(self, widget, asd):
        self.helpWindow.hide()
        return True


    # ADD DIALOG FUNCTIONS
    def addDialogApply(self, asd):
        # read
        readIP = self.addEntryIP.get_text()
        readHostname = self.addEntryHostname.get_text()
        readColor = self.addEntryColor.get_text()

        # check and append
        if ( isValidHostname(readHostname) ) or ( isValidIP(readIP) ): # if either one not empty
            if not ( isValidHTMLColor(readColor) ):
                self.statusBar.push(self.statusBarCI, "Invalid color for the new entry: Defaulted to Black")
                readColor = "black" # default: black
            appendedIter = self.devicesList.append([readHostname, readIP, False, readColor])
            if ( readIP == "" ):
                query = DNSResolveThread(self, appendedIter, readHostname, wemust.HOSTNAMETYPE)
                self.dataShare.threadsList.append(query)
                query.start()
            elif ( readHostname == "" ):
                query = DNSResolveThread(self, appendedIter, readIP, wemust.IPv4TYPE)
                query.start()
                self.dataShare.threadsList.append(query)

        self.addEntryIP.set_text("")
        self.addEntryHostname.set_text("")
        self.addEntryColor.set_text("")
        self.addDialog.hide()


    def addDialogCancel(self, asd):
        self.addDialog.hide()
        return True

    def remEntry(self, button):
        model, treeiter = self.devicesView.get_selection().get_selected()
        self.devicesList.remove(treeiter)

    def startScan(self, pippo):
        gobject.idle_add(self.scanPerform)

    def scanPerform(self):
        self.devicesList.clear()
        sT = scannerThread(self,self.STagList) # in wemust-netdbg I'm not interested in looking for specific tags, but I'll leave the chance for later implementations
        self.dataShare.threadsList.append(sT)
        sT.start() # start scanning
        time.sleep(1) # leave time for sT to start
        #self.waitPopupShow('Scanning... Please Wait')
        #waitPops = pops(self.window)
        p.show()
        self.scanLock.acquire() # blocks here until scan is finished
        #self.waitPopupHide()
        p.hide()
        while ( sT.rawReceivedStuff ): # ...if any
            self.rawReceivedStuff.append(sT.rawReceivedStuff.pop()) # cut and paste, not copy to avoid duplicates
        self.scanLock.release()
        self.Parse_Populate()


    # Parse reply and populate device list
    def Parse_Populate(self):
        self.listPeers = [] # list of tuples (hostname, IP). hostname is the SABy NID field
        colorIdx = 0
        while ( self.rawReceivedStuff ):
            #check SABY string
            popd = self.rawReceivedStuff.pop()
            addr = popd[0]
            msg = popd[1]

            if( string.find(msg, wemust.SABy_STRING) >= 0 ):
                if( string.find(msg, wemust.SABy_REPLY_STRING) >= 0 ): # is a valid reply string
                    # todo: add check for tags in self.STagList
                    # todo: add entry box for tags from the GUI
                    SyP = wemust.SAByParams()
                    SyP.setFromAnnounceMsg(msg)
                    # got nid as a hostname for a temporary append
                    appendedIter = self.devicesList.append( [SyP.nid, addr[0], True, self.dataShare.color[colorIdx] ] )
                    colorIdx = colorIdx + 1
                    # now the hostname can be fetched from DNS
                    query = DNSResolveThread(self, appendedIter, addr[0], wemust.IPv4TYPE)
                    self.dataShare.threadsList.append(query)
                    query.start()



    # SEND COMMANDS FROM STATUS or DIAGNOSTIC ENTRY TO THE TARGET(s)
    def sendCmd_TCP(self, cmd):
        ipList = []
        # hostnameList = []

        i = self.devicesList.get_iter_first()
        while i:
            if self.devicesList.get_value(i, 2) == True: # if the Address It toggle is ON
                if ( self.devicesList.get_value(i, 1) ):
                    ipList.append( self.devicesList.get_value(i, 1) ) # column 1 = IP
                elif ( self.devicesList.get_value(i, 0) ):
                    ipList.append( self.devicesList.get_value(i, 0) ) # it is ok to append a hostname, socket.connect will handle both IPs and hostnames
                    # hostnameList.append( self.devicesList.get_value(i, 0) )
                else:
                    self.statusBar.push(self.statusBarCI, "Warning: element " + i + " of the device list has no hostname and IP")


            i = self.devicesList.iter_next(i)

        for r in range(len(ipList)): # starts from 0
            destIP = ipList[r]
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((destIP, wemust.NETDBG_CTRL_PORT)) # Todo: risk of blocking here with commands such as top?
                sock.send(cmd)
                sock.close()
            except socket.error as sErr:
                self.statusBar.push(self.statusBarCI, "Warning: socket err " + str(sErr.errno) + ': ' + str(sErr.strerror) )

                # theoretically hostnames are handled as well as IPs it should be safe to delete this
                # for r in range(len(hostnameList)): # starts from 0
                #     destHostname = hostnameList[r]
                #     try:
                #         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                #         sock.connect((destHostname, NETDBG_CTRL_PORT))
                #         sock.send(cmd)
                #         sock.close()
                #     except socket.error as sErr:
                #         print("Warning: socket err " + str(sErr.errno) + ': ' + str(sErr.strerror) )



    # SEND A DIAGNOSTIC CHECK (CALLED VIA STATUS BUTTON)
    def defaultDiagnostic(self, button):
        cmd = "ps -e -o pid,uname,pcpu,pmem,comm,args | grep [j]ack"
        self.sendCmd_TCP(cmd)
        cmd = "ps -e -o pid,uname,pcpu,pmem,comm,args | grep [w]emust"
        self.sendCmd_TCP(cmd)

    ### FACTORY-DEFINED COMMANDS SEND
    def sendStart(self, button):
        self.sendCmd_TCP(STARTCMD)
    def sendStop(self, button):
        self.sendCmd_TCP(STOPCMD)
    def sendReboot(self, button):
        self.sendCmd_TCP(REBOOTCMD)
    def sendHalt(self, button):
        self.sendCmd_TCP(HALTCMD)


    ### CONSOLE CLEARING AND RELATED FUNCTIONS
    def clearConsoles(self, button):
        self.setClearSema()
    def setClearSema(self):
        for i in range(NCONSOLES):
            self.clearSema.release()
    def getClearSema(self):
        return self.clearSema
    def unsetClearSema(self):
        for i in range(NCONSOLES):
            self.clearSema.acquire()


# ------------------------------------------------------------------------ #
# ---------------- SCANNER THREAD - FINDS WEMUST NODES ------------------- #
# ------------------------------------------------------------------------ #

# This thread only retrieves raw data from possible wemust nodes, parsing is left for the GUI
class scannerThread(threading.Thread):
    SyP = wemust.SAByParams()

    def __init__(self, guiRoot, tagString):
        threading.Thread.__init__(self)
        self.guiRoot = guiRoot
        self.tagString = tagString
        self.rawReceivedStuff = []
        self.timeExpired = threading.Event()

    def run(self):
        try:
            # open up socket for listening for replies (with timeout)
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind( ('', wemust.SABy_REPLY_TCP_PORT) ) # 'None' binds to all available interfaces
            s.settimeout(SOCK_TIMEOUT) # blocks for a limited amount of time, allows to quit scanning thread
            # send SEARCH via UDP
            self.SyP.search()


            # listen for connections on TCP
            self.guiRoot.scanLock.acquire()
            now = round(time.time())
            while( not self.timeExpired.is_set() ): # we ask scannerThread to give up when time expires, GUI blocks on a dialog (how to stop this thread if it does not release the Lock?)
                if( round(time.time()) - now > SCAN_TIMEOUT - SOCK_TIMEOUT): # -SOCK_TIMEOUT takes account of the time it will take till the next while() check
                    self.timeExpired.set()
                try:
                    s.listen(wemust.WEMUST_MAX_NODES)
                    conn, addr = s.accept() # blocking with timeout
                    if conn:
                        data = conn.recv(wemust.SABy_MAX_MESSAGE_LEN) # blocking with timeout
                        if data:
                            self.rawReceivedStuff.append([addr, data]) # beware: addr is a tuple
                        conn.close()
                except socket.timeout:
                    pass
            # pass the info to the parent and finish
            while( self.rawReceivedStuff ):
                self.guiRoot.rawReceivedStuff.append(self.rawReceivedStuff.pop()) # cut and paste to parent, not copy to avoid duplicates
            self.guiRoot.scanLock.release()
        except (KeyboardInterrupt, SystemExit):
            print("Keyboard Interrupt caught: exiting")
            raise

# ------------------------------------------------------------------------ #
# ---------------- NETWORK THREADS HANDLING FUNCTIONS -------------------- #
# ------------------------------------------------------------------------ #

# SPAWN ALL NET THREADS
def threadSpawn(dataShare):


    # terminal emulator instances
    dataShare.threadsList.append( NetRxThread(dataShare.dbgQ, dataShare.globalQLock, wemust.NETDBG_PRINT_PORT, dataShare) )
    dataShare.threadsList.append( NetRxThread(dataShare.jtcliQ, dataShare.globalQLock, wemust.NETDBG_JTCLI_PORT, dataShare) )
    dataShare.threadsList.append( NetRxThread(dataShare.jdQ, dataShare.globalQLock, wemust.NETDBG_JACKD_PORT, dataShare) )
    dataShare.threadsList.append( NetRxThread(dataShare.jtserQ, dataShare.globalQLock, wemust.NETDBG_JTSER_PORT, dataShare) )

    for t in range(len(dataShare.threadsList)):
        dataShare.threadsList[t].start()

# CLOSE ALL THE THREADS
def properShutdown(self, dataShare):

    dataShare.quitFlag.set()

    print("Quit...")
    for t in range(len(dataShare.threadsList)):
        ts = time.time()
        dataShare.threadsList[t].join(QUIT_TIMEOUT)
        # if the time spent on join() is close to the timeout time it means join spent his time waiting
        # we want to track this down. We can't rely on .isAlive because it always returns true even if join
        # finished fine before the timeout
        if ( time.time() - ts > QUIT_TIMEOUT - 2):
            print("Warning on exit: thread still alive") # maybe it is locked on a resource
            print(threading.enumerate())
    Gtk.main_quit()


# ------------------------------------------------------- #
# ---------------- TERMINAL EMULATOR -------------------- #
# ------------------------------------------------------- #

# CLASS
class TermEmu():
    def __init__(self, view, guiRoot, myQ):
        #list of things not yet printed
        self.printQueue = myQ
        self.buffer = view.get_buffer()
        self.view = view
        self.clearSema = guiRoot.getClearSema()
        self.guiRoot = guiRoot

        #one thread will be adding to the print queue, and another will be iterating through it.
        #better make sure one doesn't interfere with the other.
        #self.printQueueLock = threading.Lock()
        self.printQueueLock = guiRoot.dataShare.globalQLock


    #check for new messages every self.sleepTime and clear the buffer if needed
    def update(self):
        try:
            with self.printQueueLock:
                gobject.idle_add( self.popQ ) # the only safe way to call GTK stuff from a thread
        except (KeyboardInterrupt, SystemExit):
            print("Keyboard Interrupt caught: exiting")
            raise
        if ( self.clearSema.acquire(False) ): # blocking=False
            gobject.idle_add( self.clearConsole )
            # no semaphore release needed

    # pops a line from the Rx queue and print it to console with color highlighting and scrolling
    def popQ(self):
        for i in range(len(self.printQueue)):
            msg = self.printQueue.popleft()
            addr = re.search( r'[0-9]+(?:\.[0-9]+){3}', msg ) # clumsy: I had the IP at the RxThread, but enqueueing it would be problematic. I could use a queue of structs
            found = False
            try:
                for row in range(len(self.guiRoot.devicesList)):
                    if addr.group(0) == str(self.guiRoot.devicesList[row][1]):
                        if not self.guiRoot.dataShare.printAddress: # delete address from incoming message (rely only on colors)
                            msg = msg.replace(str(addr.group(0)+ADDRESS_SEPARATOR), '', 1) # only 1st occurrence
                        self.buffer.insert_with_tags_by_name(self.buffer.get_end_iter(), msg, self.guiRoot.devicesList[row][3])
                        found = True
                        break
            except:
                pass

            if found == False:
                self.buffer.insert_with_tags_by_name(self.buffer.get_end_iter(), msg, "white")
            # scroll to bottom
            self.view.scroll_mark_onscreen( self.buffer.get_insert() )

    # CLEAR CONSOLE
    def clearConsole(self):
        self.buffer.delete(self.buffer.get_start_iter(), self.buffer.get_end_iter()) # eventuale alternativa: set_text(" ") ?
        self.buffer.remove_all_tags(self.buffer.get_start_iter(), self.buffer.get_end_iter())


# ------------------------------------------------------- #
# ---------------- DNS SEARCH THREAD -------------------- #
# ------------------------------------------------------- #

class DNSResolveThread(threading.Thread):
    def __init__(self, guiRoot, dLIter, string, type):
        threading.Thread.__init__(self)
        self.string = string
        self.type = type
        self.devicesList = guiRoot.devicesList
        self.dataShare = guiRoot.dataShare
        self.path = self.devicesList.get_path(dLIter)


    def DNSResolve(self):
        try:
            if ( self.type == wemust.HOSTNAMETYPE ):
                return socket.gethostbyname(self.string)
            elif ( self.type == wemust.IPv4TYPE):
                return socket.gethostbyaddr(self.string)
        except socket.error:
            return None


    def run_no(self):
        try:
            if ( self.type == wemust.HOSTNAMETYPE ):
                try:
                    ip = socket.gethostbyname(self.string)
                    self.devicesList[self.path][1] = ip
                    self.dataShare.toKill.append(self) # I'm done, kill me
                except socket.error:
                    self.dataShare.toKill.append(self)
                    pass

            elif ( self.type == wemust.IPv4TYPE):
                try:
                    hostname = socket.gethostbyaddr(self.string)
                    self.devicesList[self.path][0] = hostname[0]
                    self.dataShare.toKill.append(self) # I'm done, kill me
                except socket.error:
                    self.dataShare.toKill.append(self)
                    pass
        except (KeyboardInterrupt, SystemExit):
            print("Keyboard Interrupt caught: exiting")
            raise

    def run(self):
        try:
            if ( self.type == wemust.HOSTNAMETYPE ):
                ip = self.DNSResolve() # IP
                if ( ip ):
                    self.devicesList[self.path][1] = ip

            elif ( self.type == wemust.IPv4TYPE):
                addr = self.DNSResolve()
                if( addr ):
                    self.devicesList[self.path][0] = addr[0]
        except (KeyboardInterrupt, SystemExit):
            print("Keyboard Interrupt caught: exiting")
            raise





# ---------------------------------------------------------- #
# ---------------- NETWORK THREAD CLASS -------------------- #
# ---------------------------------------------------------- #

# NetRxThread: thread to receive from the network and hand it to the consoles
class NetRxThread(threading.Thread):
    data = ""
    def __init__(self, termQueue, termQueueLock, port, dataShare):
        threading.Thread.__init__(self)
        self.termQueue = termQueue
        self.termQueueLock = termQueueLock
        self.RxBufSize = 16384 # a high value can avoid assertion with large messages? gtk_text_view_validate_onscreen: assertion failed: (priv->onscreen_validated)
        self.sock = dbgUDPSockRx(0, wemust.NETDBG_MCAST_GROUP, port) # ID still useful?
        self.sock.settimeout(2) # avoid blocking calls to shut down this thread properly
        self.dataShare = dataShare


    # START THREAD
    def run(self):
        while( not self.dataShare.quitFlag.is_set() ):
            try:
                self.data, self.addr = self.sock.recvfrom(self.RxBufSize)
                with self.termQueueLock:
                    if self.dataShare.printTimestamp == True:
                        #print self.data # uncomment this line when BB messages are not shown (after switch from WLAN to DII ETH)
                        self.termQueue.append(self.addr[0] + ADDRESS_SEPARATOR + TIMESTAMP_DELIM1 + time.strftime("%H:%M:%S", time.gmtime()) + TIMESTAMP_DELIM2 + self.data + '\n') # alternative witj
                    else:
                        #print self.data # uncomment this line when BB messages are not shown (after switch from WLAN to DII ETH)
                        self.termQueue.append(self.addr[0] + ADDRESS_SEPARATOR + self.data + '\n') # Todo: translate IP into hostname
            except socket.timeout: # timeout has been set for sockets, this guarantees no exc is raised and the thread will go on listening for another timeout cycle
                pass
            except (KeyboardInterrupt, SystemExit):
                print("Keyboard Interrupt caught: exiting")
                raise


# ---------------------------------------------------------------- #
# ---------------- SOCKETS FOR NETWORK THREADS-------------------- #
# ---------------------------------------------------------------- #

class dbgUDPSockRx(socket.socket):
    dbgIP = ""
    dbgPORT = 0
    logFile = ""
    ID = 0 # when initializing it you'd better give a value >0

    def __init__(self, ID, dbgIP, dbgPORT):
        self.ID = ID
        self.dbgIP = dbgIP
        self.dbgPORT = dbgPORT

        socket.socket.__init__(self, family=socket.AF_INET, type=socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP, _sock=None)
        try:
            self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except AttributeError:
            pass # Some systems don't support SO_REUSEPORT

        try:
            self.bind((dbgIP, dbgPORT))
        except:
            print("Error: could not bind %s:%d" % (dbgIP, dbgPORT))
            self.close()
            raise

        mreq = struct.pack("=4sl", socket.inet_aton(wemust.NETDBG_MCAST_GROUP), socket.INADDR_ANY)
        try:
            self.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        except:
            print("Warning: exception: " + str(sys.exc_info()[1]) + ". Connection will not be effective") # Todo: make it a popup dialog
            pass


    def close(self):
        #unsubscribe
        self.setsockopt(socket.SOL_IP, socket.IP_DROP_MEMBERSHIP, socket.inet_aton(self.dbgIP) + socket.inet_aton('0.0.0.0'))
        self.close()


# ------------------------------------------------------------ #
# ------------------------------------------------------------ #
# --------------------- MAIN MAIN MAIN ----------------------- #
# ------------------------------------------------------------ #
# ------------------------------------------------------------ #

share = dataShare()
ndGUI = netdbgGUI(share)
ndGUI.window.show()
p = pops(ndGUI.window)
p.hide()

tS = threadSpawn(share)
tH = threadHandler(share)
tH.start()

Gtk.main()
